require 'date'
require 'yaml/store'
require 'codebreaker_game/version'
require 'codebreaker_game/random'
require 'codebreaker_game/storage'
require 'codebreaker_game/entities/user'
require 'codebreaker_game/entities/winner'
require 'codebreaker_game/entities/difficulty'
require 'codebreaker_game/entities/difficulty_factory'
require 'codebreaker_game/errors/wrong_difficulty_error'

module CodebreakerGame
  class Game
    include CodebreakerGame::Random
    extend CodebreakerGame::Storage
    attr_reader :user

    STATUS_IN_PROGRESS = 'in_process'.freeze
    STATUS_WIN = 'win'.freeze
    STATUS_LOSE = 'lose'.freeze
    STRONG_INTERSECTION = '+'.freeze
    NOT_STRONG_INTERSECTION = '-'.freeze
    SECRET_LENGTH = 4

    def initialize
      @secret_number = generate_secret(SECRET_LENGTH)
      @status = STATUS_IN_PROGRESS
      @hints = @secret_number.to_s.chars.shuffle
      @storage = CodebreakerGame::Game.new_store
      save_storage unless CodebreakerGame::Game.storage_exists?
      synchronize_storage
    end

    def win?
      @status == STATUS_WIN
    end

    def lose?
      @status == STATUS_LOSE
    end

    def create_user(name:, difficulty:)
      @user = User.new(name, difficulty)
      sample_hints(difficulty.hints)
      @user
    end

    def sample_hints(hints)
      @hints = @hints.sample(hints)
    end

    def create_difficulty(difficulty:)
      CodebreakerGame::DifficultyFactory.new.difficulty_by_name(difficulty)
    end

    def attempt(user_number)
      return compare(@secret_number, user_number) if user.attempts?

      @status = STATUS_LOSE
      nil
    end

    def sorted_winners
      @winners.sort_by do |winner|
        [winner.user.difficulty.attempts,
         winner.user.attempts_used,
         winner.user.hints_used]
      end
    end

    def hint
      user.hint if user.hints?

      @hints.pop
    end

    def save_storage
      @storage.transaction do
        @storage[:winners] = @winners || []
      end
    end

    private

    def win
      @winners << CodebreakerGame::Winner.new(user)
      @status = STATUS_WIN
    end

    def compare(secret_number, user_number)
      user.attempt
      compare_numbers(secret_number, user_number)
    end

    def compare_numbers(secret_number, user_number)
      secret_digits = secret_number.to_s.chars
      user_digits = user_number.to_s.chars
      response = strong_intersections(secret_digits, user_digits)
      response += not_strong_intersections(secret_digits, user_digits)
      win if response == STRONG_INTERSECTION * SECRET_LENGTH
      response
    end

    def strong_intersections(secret_digits, user_digits)
      response = ''
      secret_digits.each_with_index do |digit, index|
        next unless digit == user_digits[index]

        response += STRONG_INTERSECTION
        secret_digits[index] = nil
        user_digits[index] = nil
      end
      response
    end

    def not_strong_intersections(secret_digits, user_digits)
      response = ''
      secret_digits.each do |digit|
        next unless digit && user_digits.include?(digit)

        response += NOT_STRONG_INTERSECTION
        user_digits.delete_at(user_digits.index(digit))
      end
      response
    end

    def synchronize_storage
      @storage.transaction(true) do
        @winners = @storage[:winners]
      end
    end
  end
end
